import numpy as np
import random

from pypop7.optimizers.ep.ep import EP


class FEP(EP):
    """Fast Evolutionary Programming(FEP)
        Reference
        -----------
        X. Yao, Y, Liu, G. Lin
        Evolutionary Programming Made Faster
        IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 3, NO. 2, JULY 1999
        https://ieeexplore.ieee.org/abstract/document/771163
    """
    def __init__(self, problem, options):
        EP.__init__(self, problem, options)
        self.t = 1.0 / np.sqrt(2 * np.sqrt(self.ndim_problem))
        self.t1 = 1.0 / np.sqrt(2 * self.ndim_problem)
        self.q = options.get('q', 10)
        self.init_n = options.get('init_n', 3)

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        n = self.init_n * np.ones((self.n_individuals, self.ndim_problem))
        for i in range(self.n_individuals):
            x[i] = self._initialize_mean()
            y[i] = self._evaluate_fitness(x[i])
        return x, y, n

    def iterate(self, x, y, n):
        new_x = np.empty((self.n_individuals, self.ndim_problem))
        new_n = np.empty((self.n_individuals, self.ndim_problem))
        new_y = np.empty((self.n_individuals,))
        rand1 = self.rng_optimization.standard_normal(self.ndim_problem)  # gaussian normal
        rand2 = self.rng_optimization.standard_cauchy(self.ndim_problem)  # cauchy normal
        for i in range(self.n_individuals):
            for j in range(self.ndim_problem):
                new_x[i][j] = x[i][j] + n[i][j] * rand2[j]
                new_n[i][j] = n[i][j] * np.exp(self.t1 * self.rng_optimization.standard_normal(1) + self.t * rand1[j])
            new_x[i] = np.clip(new_x[i], self.lower_boundary, self.upper_boundary)
            new_y[i] = self._evaluate_fitness(new_x[i])
        new_x = np.vstack((new_x, x))
        new_y = np.hstack((new_y, y))
        new_n = np.vstack((new_x, x))
        win = np.zeros((2 * self.n_individuals,))
        list1 = list(range(0, 2 * self.n_individuals))
        for i in range(2 * self.n_individuals):
            list2 = random.sample(list1, self.q)
            for j in list2:
                if new_y[i] >= new_y[j]:
                    win[i] += 1
        order = np.argsort(win)
        for i in range(self.n_individuals):
            x[i] = new_x[order[i]]
            y[i] = new_y[order[i]]
            n[i] = new_n[order[i]]
        return x, y, n

    def optimize(self, fitness_function=None):
        fitness = EP.optimize(self, fitness_function)
        x, y, n = self.initialize()
        fitness.extend(y)
        while True:
            x, y, n = self.iterate(x, y, n)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.ep.ep import EP


class CEP(EP):
    """Classical Evolutionary Programming with self-adaptive mutation (CEP).

    References
    ----------
    Yao, X., Liu, Y. and Lin, G., 1999.
    Evolutionary programming made faster.
    IEEE Transactions on Evolutionary computation, 3(2), pp.82-102.
    https://ieeexplore.ieee.org/abstract/document/771163
    """
    def __init__(self, problem, options):
        EP.__init__(self, problem, options)
        self.sigma = options.get('sigma')  # initial global step-size
        self.q = options.get('q', 10)  # number of opponents for pairwise comparisons
        # two learning rate factors of individual step-size
        self.tau = options.get('tau', 1.0 / np.sqrt(2.0*np.sqrt(self.ndim_problem)))
        self.tau_apostrophe = options.get('tau_apostrophe', 1.0 / np.sqrt(2.0*self.ndim_problem))

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        sigmas = self.sigma*np.ones((self.n_individuals, self.ndim_problem))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y, sigmas

    def iterate(self, x, y, n):
        new_x = np.empty((self.n_individuals, self.ndim_problem))
        new_n = np.empty((self.n_individuals, self.ndim_problem))
        new_y = np.empty((self.n_individuals,))
        rand1 = self.rng_optimization.standard_normal(self.ndim_problem)
        for i in range(self.n_individuals):
            for j in range(self.ndim_problem):
                new_x[i][j] = x[i][j] + n[i][j] * rand1[j]
                new_n[i][j] = n[i][j] * np.exp(self.t1 * self.rng_optimization.standard_normal(1) + self.t * rand1[j])
            new_x[i] = np.clip(new_x[i], self.lower_boundary, self.upper_boundary)
            new_y[i] = self._evaluate_fitness(new_x[i])
        new_x = np.vstack((new_x, x))
        new_y = np.hstack((new_y, y))
        new_n = np.vstack((new_x, x))
        win = np.zeros((2 * self.n_individuals,))
        list1 = list(range(0, 2 * self.n_individuals))
        for i in range(2 * self.n_individuals):
            list2 = random.sample(list1, self.q)
            for j in list2:
                if new_y[i] >= new_y[j]:
                    win[i] += 1
        order = np.argsort(win)
        for i in range(self.n_individuals):
            x[i] = new_x[order[i]]
            y[i] = new_y[order[i]]
            n[i] = new_n[order[i]]
        return x, y, n

    def optimize(self, fitness_function=None, args=None):
        fitness = EP.optimize(self, fitness_function)
        x, y, n = self.initialize()
        fitness.extend(y)
        while True:
            x, y, n = self.iterate(x, y, n)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        return self._collect_results(fitness)



from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.res import RES
from pypop7.optimizers.es.ssaes import SSAES
from pypop7.optimizers.es.dsaes import DSAES


__all__ = [ES, RES, SSAES, DSAES]



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class RS(Optimizer):
    """Random (stochastic) Search (optimization) (RS).

    This is the **base** class for all RS classes (including local search).

    .. note:: At least its two methods (`initialize`, `iterate`) should be implemented by its subclass.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `1000`);
              and with one particular setting (`key`):
                * 'x'     - initial (starting) point (`array_like`).

    Attributes
    ----------
    x     : `array_like`
            initial (starting) point.

    Methods
    -------

    References
    ----------
    Bergstra, J. and Bengio, Y., 2012.
    Random search for hyper-parameter optimization.
    Journal of Machine Learning Research, 13(2).
    https://www.jmlr.org/papers/v13/bergstra12a.html

    Appel, M.J., Labarre, R. and Radulovic, D., 2004.
    On accelerated random search.
    SIAM Journal on Optimization, 14(3), pp.708-731.
    https://epubs.siam.org/doi/abs/10.1137/S105262340240063X

    Solis, F.J. and Wets, R.J.B., 1981.
    Minimization by random search techniques.
    Mathematics of operations research, 6(1), pp.19-30.
    https://pubsonline.informs.org/doi/abs/10.1287/moor.6.1.19

    Schrack, G. and Choit, M., 1976.
    Optimized relative step size random searches.
    Mathematical Programming, 10(1), pp.230-244.
    https://link.springer.com/article/10.1007/BF01580669

    Schumer, M.A. and Steiglitz, K., 1968.
    Adaptive step size random search.
    IEEE Transactions on Automatic Control, 13(3), pp.270-276.
    https://ieeexplore.ieee.org/abstract/document/1098903

    Matyas, J., 1965.
    Random optimization.
    Automation and Remote control, 26(2), pp.246-253.
    https://tinyurl.com/25339c4x  (in Russian)

    Rastrigin, L.A., 1963.
    The convergence of the random search method in the extremal control of a many parameter system.
    Automaton & Remote Control, 24, pp.1337-1342.
    https://tinyurl.com/djfdnpx4

    Brooks, S.H., 1958.
    A discussion of random methods for seeking maxima.
    Operations Research, 6(2), pp.244-251.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.x = options.get('x')  # initial (starting) point
        # reset its default value from 10 to 1000 (since typically more iterations are run)
        if options.get('verbose_frequency') is None:
            self.verbose_frequency = 1000
        self._n_generations = 0  # number of generations

    def initialize(self):
        raise NotImplementedError

    def iterate(self):  # for each iteration (generation)
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        is_initialization = True
        while not self._check_terminations():
            if is_initialization:
                x = self.initialize()
                is_initialization = False
            else:
                x = self.iterate()
            y = self._evaluate_fitness(x, args)
            if self.record_fitness:
                fitness.append(y)
            self._print_verbose_info(y)
            self._n_generations += 1
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class EDA(Optimizer):
    """Estimation of distribution algorithms(EDA)

probabilistic graphical models

    Based on Genetic Algorithms (GAs), the new class of algorithms generalizes GAs by replacing the crossover and mutation operators by learning and sampling the probability distribution of the best individuals of the population at each iterations of the algorithm

working in such a way, the relationships between the variables involved in the problem domain are explicitly and effectively captured and exploited

AKA probabilistic model-building genetic algorithms (PMBGAs), iterated density estimation evolutionary algorithms (IDEAs)

EDAs were introduced in the field of evolutionary
computation for the first time by Miihlenbein and PaaB (1996)

The main problem with EDAs is how the probability distribution Pl(X) is estimated.

    Reference
    --------------
    P. A. N. Bosman, D. Thierens
    Expanding from Discrete to Continuous Estimation of Distribution Algorithms: The IDEA
    PPSN 2000: 767-776

    Selection and reinforcement learning for combinatorial optimization.
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self._n_generations = 0
        if self.n_individuals is None:  # number of offspring, offspring population size (λ: lambda)
            self.n_individuals = 4 + int(3 * np.log(self.ndim_problem))  # for small population setting
        if self.n_parents is None:  # number of parents, parental population size (μ: mu)
            self.n_parents = int(self.n_individuals / 2)

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            best_so_far_y = -self.best_so_far_y if self._is_maximization else self.best_so_far_y
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, best_so_far_y, np.min(y), self.n_function_evaluations))

    def _initialize_x(self, is_restart=False):
        if is_restart or (self.x is None):
            x = self.rng_initialization.uniform(self.initial_lower_boundary,
                                                self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        return x

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['mean'] = mean
        results['sigma'] = self.sigma
        results['_n_generations'] = self._n_generations
        results['n_restart'] = self.n_restart
        return results
